/**
 * @file ts_modify.c
 *
 * This file provides an example test interface for reading/writing to
 * EMAC's version of the AT91 TSADCC Driver. Attempting to use the program
 * generated from this code on anything other than a SoM-9g45 with EMAC's
 * version of the AT91 TSADCC Driver does not make sense. Note that analogous
 * control over the driver can be accomplished using the
 * /sys/class/gpio/indexed_tsopt/index and /sys/class/gpio/indexed_tsopt/data
 * files from the GNU/Linux command line. See the README.atmel_tsadcc_sysfs file
 * for more details.
 *
 * @author EMAC, Inc. <support@emacinc.com>
 *
 * Copyright (C) 2011 EMAC, Inc.
 */
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <asm/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <termios.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include "gpio_char.h"

/*
 * see arch/arm/mach-at91/include/mach/board.h
 */
struct at91_tsadcc_data {
	int		current_index;
	unsigned int    adc_clock;
	__u32		pendet_debounce;
	__u32		ts_sample_hold_time;
	__u32		channel_enable;
	__u32		ts_pressure_threshold;
	__u32		ts_filter_samples;
	__u32		ts_x_resistances;
} curr_tsdata, prev_tsdata;

#define TS_TSSHTIM		0xffef
#define TS_PENDBC		0xfff0	
#define TS_PRESCALAR		0xfff4
#define TS_Rxp			0xfff8
#define TS_PRESSURE_THRESH	0xfffc
#define TS_FILTER_SAMPLES	0xffff

//#define INDEXED_TSOPT_PATH	"/sys/class/gpio/indexed_tsopt/"
#define INDEXED_TSOPT_PATH	"/dev/indexed_tsopt"
			
#define	INCREMENT(value, min, max, inc) \
if (value + inc < min) \
	goto value_min; \
if (value + inc > max) \
	goto value_max; \
value += inc;			

static void terminal_menu_setup(void);
static int menu(int);
static int set_data(int, __u32, __u32); 
static __u32 get_data(int, __u32); 
static int get_all(int, struct at91_tsadcc_data *); 
static int set_all(int, struct at91_tsadcc_data *); 

static char inchar;
static int stop = 0;
static char inc = 1;
static struct termios previous;

void catch_sig (int sig) 
{
	stop = 1;
}

int main(int argc, char **argv) 
{
	int fd;
	char *dev = INDEXED_TSOPT_PATH;

	/*
	 * check arguments
	 */
	if ( argc > 1 ) {
		fprintf(stderr,"%s: This program does not take arguments!", argv[0]);
		exit(EXIT_FAILURE);
	}

	/*
	 * prepare file descriptors
	 * /dev/indexed_tsopt
	 */
	if ((fd = open(dev, O_RDWR)) <=0) {
		printf("Could not open %s\n", dev);
		return -1;
	}

	/*
	 * initialize initial values
	 */
	if ((get_all(fd, &prev_tsdata) != 0) || (get_all(fd, &curr_tsdata) != 0)) {
		printf("Failed to get initial touchscreen data!");
		return(EXIT_FAILURE);
	}

	terminal_menu_setup();

	while (menu(fd)) {
		set_all(fd, &curr_tsdata);
		get_all(fd, &curr_tsdata);
		if (stop == 1)
			break;
	}

	tcsetattr(0, TCSANOW, &previous);
	return(EXIT_SUCCESS);

}

static int menu(int fd) 
{
	int retry = 1;
	int sign = 1;

	static const char *menu_items = "\n\t\tControl the Touchscreen\n"
		"For commands with two characters, the left character signifies \n"
		" an increase while the right character signifies a decrease. It\n"
		" is not necessary to press \"Enter\".\n"
		"\t( q a ) -- ADC Clock Frequency Prescalar\n"
		"\t( w s ) -- Pen Detect Debounce Register\n"
//		"\t( d e ) -- TS Sample and Hold Time\n"
		"\t( r f ) -- # of Filter Samples Per Measurement\n"
		"\t( t g ) -- Pressure Threshhold (sp?)\n"
		"\t( y h ) -- Change the amount by which values are incremented.\n"
		"\t(  1  ) -- Reset Touchscreen Values\n" 
		//"\t(  0  ) -- Save and Quit\n"
		"\t(  2  ) -- Do Not Save and Quit\n"
		"\t(Ctl-C) -- Exit Program\n";

	/*
	 * print menu
	 */

	printf("%s",menu_items);

	/*
	 * print tsadcc values
	 */
	
	printf("Initial Touchscreen Values\n");
	printf("	ADC Prescalar: %d\n "
		"	PenDBC: %d\n "
		"	Pres. Thresh: %d\n "
		"	TS Filter Samples: %d\n", \
			prev_tsdata.adc_clock, \
			prev_tsdata.pendet_debounce, \
			prev_tsdata.ts_pressure_threshold, \
			prev_tsdata.ts_filter_samples);

	printf("\nCurrent Touchscreen Values\n");
	printf("	ADC Prescalar: %d\n "
		"	PenDBC: %d\n "
		"	Pres. Thresh: %d\n "
		"	TS Filter Samples: %d\n "
		"	Increment-By: %d\n", \
			curr_tsdata.adc_clock, \
			curr_tsdata.pendet_debounce, \
			curr_tsdata.ts_pressure_threshold, \
			curr_tsdata.ts_filter_samples, \
			inc);

	fflush(stdout);

	while (retry) {
		retry = 0;
		sign = 1;

		inchar = fgetc(stdin);

		switch (tolower(inchar)) {
		case 'a':
			sign = -1;
		case 'q':
			INCREMENT(curr_tsdata.adc_clock, \
					0x4, 0xff, sign*inc)
			break;
		case 's':
			sign = -1;
		case 'w':
			INCREMENT(curr_tsdata.pendet_debounce, \
					0x1, 0xf, sign*inc)
			break;
/*
 * does not work due to treatment of TSSHTIM in kernel driver
 *
 		case 'd':
			sign = -1;
		case 'e':
			INCREMENT(curr_tsdata.ts_sample_hold_time, \
					0x1, 0xf, sign*inc)
			break;
*/
		case 'g':
			sign = -1;
		case 't':
			INCREMENT(curr_tsdata.ts_pressure_threshold, \
					0x100, 0x1000, sign*inc)
			break;
		case 'f':
			sign = -1;
		case 'r':
			INCREMENT(curr_tsdata.ts_filter_samples, \
					0x0, 0x1e, sign*inc)
			break;
		case 'h':
			sign = -1;
		case 'y':
			INCREMENT(inc, 0x1, 0xf, sign*1);
			break;			
		case '0':
			break;
			printf("Really save? Press 'y' to save, otherwise"
					" return to the menu.");
			if ((inchar = getchar()) == 'y') {
				set_all(fd, &curr_tsdata);
				return 0;
			}
			break;			
		case '1':
			printf("Reset TS values? Press 'y' to reset, otherwise"
					" return to the menu.");
			if ((inchar = getchar()) != 'y') {
				break;
			}
			curr_tsdata = prev_tsdata;
			set_all(fd, &curr_tsdata);
			get_all(fd, &curr_tsdata);
			break;
		case '2':
			curr_tsdata = prev_tsdata;
			set_all(fd, &curr_tsdata);
			get_all(fd, &curr_tsdata);
			stop = 1;
			break;
		default:
			if (stop == 1) 
				break;
			retry = 1;
			break;
		} 

		continue;

value_max:
		printf("Operation would exceed maximum allowable value.\n");
		retry = 1;
		continue;
value_min:
		printf("Operation would exceed minimum allowable value.\n");
		retry = 1;
	}

	return -1;
}

/* 
 * stolen from lcb_test.c
 */
static void terminal_menu_setup(void) 
{
	static struct termios current;
	struct sigaction act;

	printf("TSADCC Modify v.01 - EMAC, Inc. "__DATE__"\n");

	/* set up signal handler to stop each task */
	tcgetattr(0, &current);
	previous = current;
	act.sa_handler = catch_sig;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGINT, &act, 0);

	/* set up the terminal for character at a time mode */
	current.c_cc[VMIN] = 1;
	current.c_cc[VTIME] = 0;
	current.c_lflag &= ~ICANON;
	current.c_lflag &= ~ECHO;
	tcsetattr(0, TCSANOW, &current);
}

static int set_data(int fd, __u32 index, __u32 data)
{
	if (ioctl(fd, INDEXWRITE, &index)) {
		printf("Operation not supported by this device!\n");
		return -1;
	}

	if (ioctl(fd, DATAWRITE, &data)) {
		printf("Operation not supported by this device!\n");
		return -1;
	}

	return 0;
}

static __u32 get_data(int fd, __u32 index)
{
	int retval;

	if (ioctl(fd, INDEXWRITE, &index)) {
		printf("Operation not supported by this device!\n");
		return -1;
	}

	if (ioctl(fd, DATAREAD, &retval)) {
		printf("Operation not supported by this device!\n");
		return -1;
	}

	return retval;
}

static int get_all(int fd, struct at91_tsadcc_data *d) 
{
	d->adc_clock		  = get_data(fd, TS_PRESCALAR);
	d->pendet_debounce	  = get_data(fd, TS_PENDBC);
	d->ts_sample_hold_time	  = get_data(fd, TS_TSSHTIM);
	d->ts_pressure_threshold  = get_data(fd, TS_PRESSURE_THRESH);
	d->ts_filter_samples	  = get_data(fd, TS_FILTER_SAMPLES);
	d->ts_x_resistances	  = get_data(fd, TS_Rxp);

	return 0;
}

static int set_all(int fd, struct at91_tsadcc_data *d) 
{
	set_data(fd, TS_PRESCALAR, d->adc_clock);
	set_data(fd, TS_PENDBC, d->pendet_debounce);
	set_data(fd, TS_TSSHTIM, d->ts_sample_hold_time);
	set_data(fd, TS_PRESSURE_THRESH, d->ts_pressure_threshold);
	set_data(fd, TS_FILTER_SAMPLES, d->ts_filter_samples);
	set_data(fd, TS_Rxp, d->ts_x_resistances);

	return 0;
}
